__all__ = [
    "media_extensions", "image_extensions",
    "get_anchor_domain", "get_anchor_scheme",
    "INITIAL_STATIC_TOKEN_SIZE", "STATIC_TOKEN_CHOICES",
    "default_uctoken_duration", "TOKEN_SIZE", "FILE_TOKEN_SIZE",
    "force_captcha", "VALID_INTENTIONS", "VALID_SUB_INTENTIONS",
    "get_requests_params"
]
import re
import functools
import datetime
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django.shortcuts import resolve_url

from spkcspider.constants import host_tld_matcher
from spkcspider.utils.settings import get_settings_func

# webbrowser supported image formats
image_extensions = set(getattr(
    settings, "SPIDER_IMAGE_EXTENSIONS", {
        "jpg", "jpeg", "bmp", "png", "ico", "svg", "gif", "webp"
    }
))

media_extensions = set(getattr(
    settings, "SPIDER_MEDIA_EXTENSIONS", {
        "mp4", "ogg", "flac", "mp3", "webm", "wav", "avi"
    }
))


@functools.lru_cache(1)
def get_anchor_domain():
    # if None, set to default Site ID if models are ready
    _anchor_domain = getattr(
        settings, "SPIDER_ANCHOR_DOMAIN", None
    )
    if _anchor_domain:
        return resolve_url(_anchor_domain)
    from django.contrib.sites.models import Site
    return Site.objects.get(id=settings.SITE_ID).domain


@functools.lru_cache(1)
def get_anchor_scheme():
    if getattr(settings, "SPIDER_ANCHOR_SCHEME", None) is not None:
        return settings.SPIDER_ANCHOR_SCHEME
    _anchor_domain = get_anchor_domain()
    if any(
        re.search(pattern, _anchor_domain)
        for pattern in getattr(settings, "SECURE_REDIRECT_EXEMPT", [])
    ):
        return "http"
    return "https"


INITIAL_STATIC_TOKEN_SIZE = str(
    getattr(settings, "SPIDER_INITIAL_STATIC_TOKEN_SIZE", 12)
)

STATIC_TOKEN_CHOICES = [
    ("", ""),
    (INITIAL_STATIC_TOKEN_SIZE, _("default ({} Bytes)")),
    ("3", _("low ({} Bytes)")),
    ("12", _("medium ({} Bytes)")),
    ("30", _("high ({} Bytes)")),
]

VALID_INTENTIONS = set(getattr(
    settings, "SPIDER_VALID_INTENTIONS",
    {"auth", "domain", "live", "login", "payment", "persist", "sl"}
))
VALID_SUB_INTENTIONS = {"sl", "live"}

TOKEN_SIZE = getattr(settings, "TOKEN_SIZE", 30)
FILE_TOKEN_SIZE = getattr(settings, "FILE_TOKEN_SIZE", TOKEN_SIZE)

force_captcha = getattr(settings, "REQUIRE_LOGIN_CAPTCHA", False)

# require USE_CAPTCHAS
force_captcha = (force_captcha and getattr(settings, "USE_CAPTCHAS", False))


default_uctoken_duration = getattr(
    settings, "DEFAULT_UCTOKEN_DURATION",
    datetime.timedelta(days=7)
)


def get_requests_params(url):
    _url = host_tld_matcher.match(url)
    if not _url:
        raise ValidationError(
            _("Invalid URL: \"%(url)s\""),
            code="invalid_url",
            params={"url": url}
        )
    _url = _url.groupdict()
    mapper = settings.SPIDER_REQUEST_KWARGS_MAP
    return (
        mapper.get(
            _url["host"],
            mapper.get(
                _url["tld"],  # maybe None but then fall to retrieval 3
                mapper[b"default"]
            )
        ),
        get_settings_func(
            "SPIDER_INLINE",
            "spkcspider.apps.spider.functions.clean_spider_inline",
            exclude=frozenset({True})
        )(_url["host"])
    )



__all__ = [
    "MAX_TOKEN_SIZE", "MAX_TOKEN_B64_SIZE", "MIN_PROTECTION_STRENGTH_LOGIN",
    "ProtectionType", "ProtectionStateType", "travel_scrypt_params"
]

import enum


MAX_TOKEN_SIZE = 90

if MAX_TOKEN_SIZE % 3 != 0:
    raise Exception("MAX_TOKEN_SIZE must be multiple of 3")

MAX_TOKEN_B64_SIZE = MAX_TOKEN_SIZE*4//3


MIN_PROTECTION_STRENGTH_LOGIN = 2


travel_scrypt_params = {
    "length": 32,
    "n": 2**14,
    "r": 16,
    "p": 2
}


class ProtectionType(str, enum.Enum):
    # receives: request, scope
    access_control = "a"
    # receives: request, scope, password
    authentication = "b"
    # c: former reliable, no meaning
    # protections which does not contribute to required_passes (404 only)
    no_count = "d"
    # protections which have side effects
    side_effects = "e"
    # show password dialog
    password = "f"
    # forget about recovery, every recovery method is authentication
    # and will be misused this way
    # The only real recovery is by staff and only if there is a secret


class ProtectionStateType(str, enum.Enum):
    disabled = "a"
    enabled = "b"
    instant_fail = "c"

    @classmethod
    def as_choices(cls):
        from django.utils.translation import gettext_lazy as _
        return (
            (i[1].value, _(i[0]))
            for i in ProtectionStateType.__members__.items()
        )



# Generated by Django 2.2.4 on 2019-08-17 18:14

from django.db import migrations

from spkcspider.constants import ProtectionStateType


def forward(apps, schema_editor):
    AssignedProtection = apps.get_model("spider_base", "AssignedProtection")

    for i in AssignedProtection.objects.all():
        if i.active and i.instant_fail:
            i.state = ProtectionStateType.instant_fail.value
        elif i.active:
            i.state = ProtectionStateType.enabled.value
        else:
            i.state = ProtectionStateType.disabled.value
        i.save(update_fields=["state"])


def backward(apps, schema_editor):
    AssignedProtection = apps.get_model("spider_base", "AssignedProtection")

    for i in AssignedProtection.objects.all():
        if i.state == ProtectionStateType.instant_fail.value:
            i.active = True
            i.instant_fail = True
        elif i.state == ProtectionStateType.enabled.value:
            i.active = True
            i.instant_fail = False
        else:
            i.active = False
            i.instant_fail = False
        i.save(update_fields=["active", "instant_fail"])


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0006_auto_20190817_1814'),
    ]

    operations = [
        migrations.RunPython(forward, backward),
    ]



__all__ = (
    "literalize", "field_to_python", "add_property", "add_by_field",
    "prepare_description"
)

import re
from urllib.parse import urljoin

from rdflib import Literal, BNode, XSD, RDF, URIRef

from django.conf import settings
from django.forms import BoundField, Field


from spkcspider.constants import spkcgraph

# for not spamming sets
_empty_set = frozenset()


def field_to_python(value):
    if isinstance(value, BoundField):
        data = value.initial
        if value.form.is_bound:
            value = value.field.bound_data(value.data, data)
        else:
            value = data
    return value


def literalize(
    ob=None, datatype=None, use_uriref=None, domain_base=""
):
    if isinstance(ob, BoundField):
        if not datatype:
            datatype = getattr(ob.field, "spkc_datatype", None)
        if use_uriref is None:
            use_uriref = getattr(ob.field, "spkc_use_uriref", None)
        ob = field_to_python(ob)
    elif isinstance(datatype, BoundField):
        if use_uriref is None:
            use_uriref = getattr(datatype.field, "spkc_use_uriref", None)
        datatype = getattr(datatype.field, "spkc_datatype", None)
    elif isinstance(datatype, Field):
        if use_uriref is None:
            use_uriref = getattr(datatype, "spkc_use_uriref", None)
        datatype = getattr(datatype, "spkc_datatype", None)
    if ob is None:
        return RDF.nil
    if hasattr(ob, "get_absolute_url"):
        if not datatype:
            datatype = spkcgraph["hashableURI"]
        if use_uriref is None:
            use_uriref = True
        ob = ob.get_absolute_url()
    elif isinstance(ob, str) and not datatype:
        datatype = XSD.string
    if use_uriref:
        return URIRef(urljoin(domain_base, ob))
    return Literal(ob, datatype=datatype)


def add_property(
    graph, name, ref=None, ob=None, literal=None, datatype=None,
    iterate=False
):
    value_node = BNode()
    if ref:
        graph.add((
            ref, spkcgraph["properties"],
            value_node
        ))
    graph.add((
        value_node, spkcgraph["name"],
        Literal(name, datatype=XSD.string)
    ))
    if literal is None:
        literal = getattr(ob, name)
    if not iterate:
        literal = [literal]
    for l in literal:
        graph.add((
            value_node, spkcgraph["value"],
            Literal(l, datatype=datatype)
        ))
    if not literal:
        graph.set((value_node, spkcgraph["value"], RDF.nil))
    return value_node


def add_by_field(dic, field="__name__"):
    def _func(klass):
        # should be able to block real object
        if "{}.{}".format(klass.__module__, klass.__qualname__) not in getattr(
            settings, "SPIDER_BLACKLISTED_MODULES", _empty_set
        ):
            dic[getattr(klass, field)] = klass

        return klass
    return _func


_cleanstr = re.compile(r'<+.*>+')
_whitespsplit = re.compile(r'\s+')


def prepare_description(raw_html, amount=0):
    text = _cleanstr.sub(' ', raw_html).\
        replace("<", " ").replace(">", " ").strip()
    return _whitespsplit.split(text, amount)



__all__ = (
    "merge_get_url", "extract_host"
)

import re
from urllib.parse import urlsplit, urlunsplit, parse_qs, urlencode

from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


_check_scheme = re.compile(r'^[a-z]+://', re.I)


def extract_host(url):
    url = url.lstrip(":/")
    if _check_scheme.search(url) is None:
        urlparsed = urlsplit("://".join(("https", url)))
    else:
        urlparsed = urlsplit(url)
    return "://".join(urlparsed[:2])


def merge_get_url(_url, **kwargs):
    if not _url.isprintable():
        raise ValidationError(
            _("Url contains control characters"),
            code="control_characters"
        )
    _url = _url.lstrip(":/")
    if _check_scheme.search(_url) is None:
        urlparsed = urlsplit("://".join(("https", _url)))
    else:
        urlparsed = urlsplit(_url)
    _strip = []
    for i in kwargs.keys():
        if not kwargs[i]:
            _strip.append(i)
    GET = parse_qs(urlparsed.query)
    GET.update(kwargs)
    for item in _strip:
        GET.pop(item, None)
    ret = urlunsplit((*urlparsed[:3], urlencode(GET, doseq=True), ""))
    return ret



# Generated by Django 2.2.4 on 2019-08-17 18:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0007_auto_20190817_1814'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='assignedprotection',
            name='active',
        ),
        migrations.RemoveField(
            model_name='assignedprotection',
            name='instant_fail',
        ),
    ]



__all__ = (
    "get_hashob", "aesgcm_scrypt_cryptor", "aesgcm_pbkdf2_cryptor",
    "calculate_protection_strength", "create_b64_token", "create_b64_id_token"
)

import logging
import base64
import os


from spkcspider.constants import MAX_TOKEN_SIZE

from hashlib import pbkdf2_hmac
from statistics import mean


from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from django.conf import settings
from django.utils.encoding import force_bytes
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from spkcspider.constants.protections import ProtectionStateType

logger = logging.getLogger(__name__)

_pbkdf2_params = {
    "iterations": 120000,
    "hash_name": "sha512",
    "dklen": 32
}

_Scrypt_params = {
    "length": 32,
    "n": 2**14,
    "r": 16,
    "p": 2
}


def get_hashob():
    return hashes.Hash(
        settings.SPIDER_HASH_ALGORITHM, backend=default_backend()
    )


def aesgcm_scrypt_cryptor(pw, salt=None, params=_Scrypt_params):
    if salt is None:
        salt = settings.SECRET_KEY
    salt = force_bytes(salt)

    return AESGCM(
        Scrypt(
            salt=salt,
            backend=default_backend(),
            **params
        ).derive(pw[:128].encode("utf-8"))
    )


def aesgcm_pbkdf2_cryptor(pw, salt=None, params=_pbkdf2_params):
    if salt is None:
        salt = settings.SECRET_KEY
    salt = force_bytes(salt)

    return AESGCM(
        pbkdf2_hmac(
            password=pw[:128].encode("utf-8"),
            salt=salt,
            **params
        )
    )


def calculate_protection_strength(required_passes, protections=None):
    prot_strength = None
    max_prot_strength = 0
    if protections:
        # instant fail strength
        fail_strength = 0
        amount = 0
        # regular protections strength
        strengths = []
        for protection in protections:
            state = protection.cleaned_data.get(
                "state", ProtectionStateType.disabled
            )
            if state == ProtectionStateType.disabled:
                continue
            if not protection.is_valid():
                continue

            if len(str(protection.cleaned_data)) > 100000:
                raise ValidationError(
                    _("Protection >100 kb: %(name)s"),
                    params={"name": protection}
                )
            if state == ProtectionStateType.instant_fail:
                s = protection.get_strength()
                if s[1] > max_prot_strength:
                    max_prot_strength = s[1]
                fail_strength = max(
                    fail_strength, s[0]
                )
            else:
                s = protection.get_strength()
                if s[1] > max_prot_strength:
                    max_prot_strength = s[1]
                strengths.append(s[0])
            amount += 1
        strengths.sort()
        if required_passes > 0:
            if amount == 0:
                # login or token only
                # not 10 because 10 is also used for uniqueness
                prot_strength = 4
            else:
                # avg strength but maximal 3,
                # (4 can appear because of component auth)
                # clamp
                prot_strength = min(round(mean(
                    strengths[:required_passes]
                )), 3)
        else:
            prot_strength = 0
        prot_strength = max(prot_strength, fail_strength)
    return (prot_strength, max_prot_strength)


def create_b64_token(size=None):
    if not size:
        from django.conf import settings
        size = getattr(settings, "TOKEN_SIZE", 30)
    if size > MAX_TOKEN_SIZE:
        logger.warning("Nonce too big")
    return base64.urlsafe_b64encode(
        os.urandom(size)
    ).decode('ascii').rstrip("=")


def create_b64_id_token(id, sep="_", size=None):
    return sep.join((hex(id)[2:], create_b64_token(size)))
